import numpy as np
from scipy import sparse
import scipy.stats.stats as sss
from scipy.sparse.linalg import lgmres, eigsh
import trimesh

########################
# error tolerance for lgmres solver
solver_tolerance = 1e-6
########################


def mesh_laplacian_eigenvectors(mesh, nb_vectors=1):
    """
    compute the nb_vectors first non-null eigenvectors of the graph Laplacian
     of mesh
    :param mesh:
    :param nb_vectors:
    :return:
    """
    lap, lap_b = compute_mesh_laplacian(mesh, lap_type="fem")
    w, v = eigsh(lap.tocsr(), nb_vectors +
                 1, M=lap_b.tocsr(), sigma=solver_tolerance)
    return v[:, 1:]


# def mesh_fiedler_length(mesh, dist_type='geodesic', fiedler=None):
#     """
#     distance between the two vertices corresponding to the min and max of
# the 2d laplacien eigen vector
#     :param mesh:
#     :param dist_type:
#     :param fiedler:
#     :return:
#     """
#     if fiedler is None:
#         fiedler = mesh_laplacian_eigenVectors(mesh, 1)
#     imin = fiedler.argmin()
#     imax = fiedler.argmax()
#     vert = np.array(mesh.vertex())
#
#     if dist_type == 'geodesic':
#         print('Computing GEODESIC distance between the max and min')
#         g = aims.GeodesicPath(mesh, 3, 0)
#         dist = g.shortestPath_1_1_len(int(imin), int(imax))
#
#     else:
#         print('Computing EUCLIDIAN distance between the max and min')
#         min_max = vert[imin, :]-vert[imax, :]
#         dist = np.sqrt(np.sum(min_max * min_max, 0))
#     return(dist, fiedler)


def laplacian_mesh_smoothing(mesh, nb_iter, dt, volume_preservation=False):
    """
    smoothing the mesh by solving the heat equation using fem Laplacian
    ADD REF
    :param mesh:
    :param nb_iter:
    :param dt:
    :return:
    """
    print("    Smoothing mesh")
    lap, lap_b = compute_mesh_laplacian(mesh, lap_type="fem")
    smoothed_vert = laplacian_smoothing(mesh.vertices, lap, lap_b, nb_iter, dt)
    if volume_preservation:
        vol_ini = mesh.volume
        vol_new = trimesh.triangles.mass_properties(
            smoothed_vert[mesh.faces], skip_inertia=True
        )["volume"]
        # scale by volume ratio
        smoothed_vert *= (vol_ini / vol_new) ** (1.0 / 3.0)
    return trimesh.Trimesh(
        faces=mesh.faces,
        vertices=smoothed_vert,
        metadata=mesh.metadata,
        process=False
    )


def laplacian_texture_smoothing(mesh, tex, nb_iter, dt):
    """
    smoothing the texture by solving the heat equation using fem Laplacian
    :param mesh:
    :param tex:
    :param nb_iter:
    :param dt:
    :return:
    """
    print("    Smoothing texture")
    lap, lap_b = compute_mesh_laplacian(mesh, lap_type="fem")
    return laplacian_smoothing(tex, lap, lap_b, nb_iter, dt)


def laplacian_smoothing(texture_data, lap, lap_b, nb_iter, dt):
    """
    sub-function for smoothing using fem Laplacian
    :param texture_data:
    :param lap:
    :param lap_b:
    :param nb_iter:
    :param dt:
    :return:
    """
    mod = 1
    if nb_iter > 10:
        mod = 10
    if nb_iter > 100:
        mod = 100
    if nb_iter > 1000:
        mod = 1000
    # print(tex.shape[0])
    # print(tex.ndim)
    # if tex.ndim < 2:
    #    Mtex = tex.reshape(tex.shape[0],1)
    # else:
    #    Mtex = tex
    # using Implicit scheme
    # B(X^(n+1)-X^n)/dt+L(X^(n+1))=0
    M = lap_b + dt * lap
    for i in range(nb_iter):
        texture_data = lap_b * texture_data
        if texture_data.ndim > 1:
            for d in range(texture_data.shape[1]):
                texture_data[:, d], infos = lgmres(
                    M.tocsr(), texture_data[:, d], rtol=solver_tolerance
                )
        else:
            texture_data, infos = lgmres(
                M.tocsr(), texture_data, rtol=solver_tolerance)
        if i % mod == 0:
            print(i)

    # using Explicit scheme, convergence guaranteed only for dt<1 and not
    # faster than implicit when using fem Laplacian
    # B(X^(n+1)-X^n)/dt+L(X^n)=0
    # M = B-dt*L
    # for i in range(Niter):
    #     Mtex = M * Mtex
    #     Mtex, infos = lgmres(B.tocsr(), Mtex, rtol=solver_tolerance)
    #     if (i % mod == 0):
    #         print(i)
    print("    OK")
    return texture_data


def compute_mesh_weights(
    mesh, weight_type="conformal", cot_threshold=None, z_threshold=None
):
    """
    compute a weight matrix
    W is sparse weight matrix and W(i,j) = 0 is vertex i and vertex j are not
    connected in the mesh.

    details are presented in:
    Desbrun, M., Meyer, M., & Alliez, P. (2002).
    Intrinsic parameterizations of surface meshes.
    Computer Graphics Forum, 21(3), 209–218.
    https://doi.org/10.1111/1467-8659.00580

    and
    Reuter, M., Biasotti, S., & Giorgi, D. (2009).
    Discrete Laplace–Beltrami operators for shape analysis and segmentation.
    Computers & …, 33(3), 381–390.
    https://doi.org/10.1016/j.cag.2009.03.005

    additional checks and thresholds are applied to ensure finite values

    :param mesh:
    :param weight_type: choice across conformal, fem, meanvalue, authalic
    :param cot_threshold:
    :param z_threshold:
    :return:
    """
    #    cot_threshold=0.00001
    #   print('angle threshold')
    print("    Computing mesh weights of type " + weight_type)
    vert = mesh.vertices
    poly = mesh.faces

    Nbv = vert.shape[0]
    W = sparse.lil_matrix((Nbv, Nbv))
    femB = sparse.lil_matrix((Nbv, Nbv))
    if weight_type == "conformal" or weight_type == "fem":
        threshold = 0.0001  # np.spacing(1)??
        threshold_needed = 0
        for i in range(3):
            i1 = np.mod(i, 3)
            i2 = np.mod(i + 1, 3)
            i3 = np.mod(i + 2, 3)
            pp = vert[poly[:, i2], :] - vert[poly[:, i1], :]
            qq = vert[poly[:, i3], :] - vert[poly[:, i1], :]
            cr = np.cross(pp, qq)
            area = np.sqrt(np.sum(np.power(cr, 2), 1)) / 2
            #             nopp = np.apply_along_axis(np.linalg.norm, 1, pp)
            #             noqq = np.apply_along_axis(np.linalg.norm, 1, qq)
            noqq = np.sqrt(np.sum(qq * qq, 1))
            nopp = np.sqrt(np.sum(pp * pp, 1))
            thersh_nopp = np.where(nopp < threshold)[0]
            thersh_noqq = np.where(noqq < threshold)[0]
            if len(thersh_nopp) > 0:
                nopp[thersh_nopp] = threshold
                threshold_needed += len(thersh_nopp)
            if len(thersh_noqq) > 0:
                noqq[thersh_noqq] = threshold
                threshold_needed += len(thersh_noqq)
            #        print(np.min(noqq))
            pp = pp / np.vstack((nopp, np.vstack((nopp, nopp)))).transpose()
            qq = qq / np.vstack((noqq, np.vstack((noqq, noqq)))).transpose()
            ang = np.arccos(np.sum(pp * qq, 1))
            # ############## preventing infs in weights
            inds_zeros = np.where(ang == 0)[0]
            ang[inds_zeros] = threshold
            threshold_needed_angle = len(inds_zeros)
            ################################
            cot = 1 / np.tan(ang)
            if cot_threshold is not None:
                thresh_inds = cot < 0
                cot[thresh_inds] = cot_threshold
                threshold_needed_angle += np.count_nonzero(thresh_inds)
            W = W + sparse.coo_matrix(
                (cot, (poly[:, i2], poly[:, i3])), shape=(Nbv, Nbv)
            )
            W = W + sparse.coo_matrix(
                (cot, (poly[:, i3], poly[:, i2])), shape=(Nbv, Nbv)
            )
            femB = femB + sparse.coo_matrix(
                (area / 12, (poly[:, i2], poly[:, i3])), shape=(Nbv, Nbv)
            )
            femB = femB + sparse.coo_matrix(
                (area / 12, (poly[:, i3], poly[:, i2])), shape=(Nbv, Nbv)
            )

        # if weight_type == 'fem' :
        #     W.data = W.data/2

        nnz = W.nnz
        if z_threshold is not None:
            z_weights = sss.zscore(W.data)
            inds_out = np.where(np.abs(z_weights) > z_threshold)[0]
            W.data[inds_out] = np.mean(W.data)
            print(
                "    -Zscore threshold needed for ",
                len(inds_out),
                " values = ",
                100 * len(inds_out) / nnz,
                " %",
            )
            # inds_out_inf = np.where(z_weights < -z_thresh)[0]
            # inds_out_sup = np.where(z_weights > z_thresh)[0]
            # val_inf = np.max(W.data[inds_out_inf])
            # W.data[inds_out_inf] = val_inf
            # val_sup = np.min(W.data[inds_out_sup])
            # W.data[inds_out_sup] = val_sup
            # print('    -Zscore threshold needed for ',
            # len(inds_out_inf)+len(inds_out_sup),' values-')
        print(
            "    -edge length threshold needed for ",
            threshold_needed,
            " values = ",
            100 * threshold_needed / nnz,
            " %",
        )
        if cot_threshold is not None:
            print(
                "    -cot threshold needed for ",
                threshold_needed_angle,
                " values = ",
                100 * threshold_needed_angle / nnz,
                " %",
            )

    if weight_type == "meanvalue":
        for i in range(3):
            i1 = np.mod(i, 3)
            i2 = np.mod(i + 1, 3)
            i3 = np.mod(i + 2, 3)
            pp = vert[poly[:, i2], :] - vert[poly[:, i1], :]
            qq = vert[poly[:, i3], :] - vert[poly[:, i1], :]
            rr = vert[poly[:, i2], :] - vert[poly[:, i3], :]
            # normalize the vectors
            noqq = np.sqrt(np.sum(qq * qq, 1))
            nopp = np.sqrt(np.sum(pp * pp, 1))
            norr = np.sqrt(np.sum(rr * rr, 1))
            pp = pp / np.vstack((nopp, np.vstack((nopp, nopp)))).transpose()
            qq = qq / np.vstack((noqq, np.vstack((noqq, noqq)))).transpose()
            rr = rr / np.vstack((norr, np.vstack((norr, norr)))).transpose()
            # compute angles
            angi1 = np.arccos(np.sum(pp * qq, 1)) / 2
            qq = -qq
            angi2 = np.arccos(np.sum(rr * qq, 1)) / 2
            W = W + sparse.coo_matrix(
                (np.tan(angi1)
                 / norr, (poly[:, i1], poly[:, i3])), shape=(Nbv, Nbv)
            )
            W = W + sparse.coo_matrix(
                (np.tan(angi2)
                 / norr, (poly[:, i3], poly[:, i1])), shape=(Nbv, Nbv)
            )
        nnz = W.nnz
    if weight_type == "authalic":
        for i in range(3):
            i1 = np.mod(i, 3)
            i2 = np.mod(i + 1, 3)
            i3 = np.mod(i + 2, 3)
            pp = vert[poly[:, i2], :] - vert[poly[:, i1], :]
            qq = vert[poly[:, i3], :] - vert[poly[:, i1], :]
            rr = vert[poly[:, i2], :] - vert[poly[:, i3], :]
            # normalize the vectors
            noqq = np.sqrt(np.sum(qq * qq, 1))
            nopp = np.sqrt(np.sum(pp * pp, 1))
            norr = np.sqrt(np.sum(rr * rr, 1))
            pp = pp / np.vstack((nopp, np.vstack((nopp, nopp)))).transpose()
            qq = qq / np.vstack((noqq, np.vstack((noqq, noqq)))).transpose()
            rr = rr / np.vstack((norr, np.vstack((norr, norr)))).transpose()
            # compute angles
            angi1 = np.arccos(np.sum(pp * qq, 1)) / 2
            cot1 = 1 / np.tan(angi1)
            qq = -qq
            angi2 = np.arccos(np.sum(rr * qq, 1)) / 2
            cot2 = 1 / np.tan(angi2)
            W = W + sparse.coo_matrix(
                (cot1 / norr**2, (poly[:, i3], poly[:, i1])), shape=(Nbv, Nbv)
            )
            W = W + sparse.coo_matrix(
                (cot2 / norr**2, (poly[:, i1], poly[:, i3])), shape=(Nbv, Nbv)
            )
        nnz = W.nnz
    li = np.hstack(W.data)
    nb_Nan = len(np.where(np.isnan(li))[0])
    nb_neg = len(np.where(li < 0)[0])
    print(
        "    -number of Nan in weights: ",
        nb_Nan,
        " = ",
        100 *
        nb_Nan /
        nnz,
        " %")
    print(
        "    -number of Negative values in weights: ",
        nb_neg,
        " = ",
        100 * nb_neg / nnz,
        " %",
    )

    return W.tocsr(), femB.tocsr()


def compute_mesh_laplacian(
        mesh, weights=None, fem_b=None, lap_type="conformal"):
    """
    compute laplacian of a mesh
    see compute_mesh_weight for details
    :param mesh:
    :param weights:
    :param fem_b:
    :param lap_type:
    :return:
    """
    print("  Computing Laplacian")
    if weights is None:
        (weights, fem_b) = compute_mesh_weights(mesh, weight_type=lap_type)

    if lap_type == "fem":
        weights.data = weights.data / 2

    N = weights.shape[0]
    sB = fem_b.sum(axis=0)
    diaB = sparse.dia_matrix((sB, 0), shape=(N, N))
    B = sparse.lil_matrix(diaB + fem_b)
    s = weights.sum(axis=0)
    dia = sparse.dia_matrix((s, 0), shape=(N, N))
    L = sparse.lil_matrix(dia - weights)

    # if symmetrize == 1 & & normalize == 0
    #     L = diag(sum(W, 2)) - W;
    # elseif
    # symmetrize == 1 & & normalize == 1
    # L = speye(n) - diag(sum(W, 2). ^ (-1 / 2)) * W * diag(
    #     sum(W, 2). ^ (-1 / 2));
    # elseif
    # symmetrize == 0 & & normalize == 1
    # L = speye(n) - diag(sum(W, 2). ^ (-1)) * W;

    li = np.hstack(L.data)
    print("    -nb Nan in Laplacian : ", len(np.where(np.isnan(li))[0]))
    print("    -nb Inf in Laplacian : ", len(np.where(np.isinf(li))[0]))

    return L, B


def depth_potential_function(mesh, curvature, alphas):
    """
    compute the depth potential function of a mesh as desribed in
    Boucher, M., Whitesides, S., & Evans, A. (2009).
    Depth potential function for folding pattern representation,
    registration and analysis.
    Medical Image Analysis, 13(2), 203–14.
    doi:10.1016/j.media.2008.09.001
    :param mesh:
    :param curvature:
    :param alphas:
    :return:
    """
    L, LB = compute_mesh_laplacian(mesh, lap_type="fem")
    B = (
        -2
        * LB
        * (curvature - (np.sum(curvature * LB.diagonal())
                        / np.sum(LB.diagonal())))
    )
    # be careful with factor 2 used in eq (13)

    dpf = []
    for ind, alpha in enumerate(alphas):
        M = alpha * LB + L / 2
        dpf_t, info = lgmres(M.tocsr(), B, rtol=solver_tolerance)
        dpf.append(dpf_t)

    ############################
    # old, slower and less accurate implementation using conformal laplacian
    # instead of fem
    ############################
    # vert_voronoi = vertexVoronoi(mesh)
    # L, LB = compute_mesh_laplacian(mesh, lap_type='conformal')
    # B = -2 * vert_voronoi * (curvature-( np.sum(curvature*vert_voronoi)
    # / vert_voronoi.sum() ))
    # B=B.squeeze()
    # for ind, alpha in enumerate(alphas):
    #     A = sparse.dia_matrix((alpha*vert_voronoi, 0), shape=(Nbv, Nbv))
    #     M = A+L
    #     dpf_t, info = lgmres(M.tocsr(), B, rtol=solver_tolerance)
    #     dpf.append(dpf_t)
    return dpf


def triangle_gradient(mesh, texture_array):
    """
    Compute gradient on a triangular mesh with a scalar function.
    Gradient is computed on each triangle by the function described in
    http://dgd.service.tu-berlin.de/wordpress/vismathws10/2012/10/
    17/gradient-of-scalar-functions/.
    first version author: Guillaume Vicaigne (Internship 2018)
    :param mesh: Triangular mesh
    :param texture_array: Scalar function on Vertices, numpy array
    :return: Gradient on Triangle
    :rtype: Matrix of size number of polygons x 3
    """

    # Initialize Parameters
    vert = mesh.vertices
    poly = mesh.faces
    l_poly = len(poly)
    n = 0
    dicgrad = np.zeros([l_poly, 3])

    # Calculate the Gradient
    for i in range(l_poly):

        # Percentage done
        if int(i / float(l_poly) * 100) > n:
            n += 1
            print(str(n) + " %")
        j = []
        for jj in poly[i]:
            j.append(jj)
        eij = [
            vert[j[1]][0] - vert[j[0]][0],
            vert[j[1]][1] - vert[j[0]][1],
            vert[j[1]][2] - vert[j[0]][2],
        ]
        eki = [
            vert[j[0]][0] - vert[j[2]][0],
            vert[j[0]][1] - vert[j[2]][1],
            vert[j[0]][2] - vert[j[2]][2],
        ]
        ejk = [
            vert[j[2]][0] - vert[j[1]][0],
            vert[j[2]][1] - vert[j[1]][1],
            vert[j[2]][2] - vert[j[1]][2],
        ]
        A = 0.5 * np.linalg.norm(np.cross(eij, ejk))
        N = 0.5 / A * np.cross(ejk, eki)
        dicgrad[i] = np.cross(
            0.5 * N / A,
            np.multiply(texture_array[j[0]], ejk)
            + np.multiply(texture_array[j[1]], eki)
            + np.multiply(texture_array[j[2]], eij),
        )

    return dicgrad


def cross_product(vec1, vec2):
    if vec1.shape != vec2.shape:
        raise Exception("Not the same size")

    res = np.zeros(vec1.shape)
    res[:, 0] = vec1[:, 1] * vec2[:, 2] - vec1[:, 2] * vec2[:, 1]
    res[:, 1] = vec1[:, 2] * vec2[:, 0] - vec1[:, 0] * vec2[:, 2]
    res[:, 2] = vec1[:, 0] * vec2[:, 1] - vec1[:, 1] * vec2[:, 0]

    return res


def gradient_fast(mesh, texture_array):
    """
    Compute gradient on a triangular mesh with a scalar function.
    Gradient is computed on each triangle by the function described in
    http://dgd.service.tu-berlin.de/wordpress/vismathws10/2012/10/
    17/gradient-of-scalar-functions/.

    Formula for the triangle
    grad(f) = - (1/2A) N x (f_i e_{jk} + f_j e_{ik} + f_k e_{ij} )
    And for a mesh
    grad(f) = 1/nb_neighbours * sum(grad_f on each triangle)

    Faster version by using numpy (J Lefevre)
    :param mesh: Triangular mesh
    :param texture_array: Scalar function on Vertices, numpy array
    :return: Gradient on Vertices
    :rtype: numpy.array
    """
    n_tri = mesh.faces.shape[0]
    n_vertex = mesh.vertices.shape[0]
    texture = np.reshape(texture_array, (n_vertex, 1))

    e_ij = mesh.vertices[mesh.faces[:, 1], :] - \
        mesh.vertices[mesh.faces[:, 0], :]
    e_ki = mesh.vertices[mesh.faces[:, 0], :] - \
        mesh.vertices[mesh.faces[:, 2], :]
    e_jk = mesh.vertices[mesh.faces[:, 2], :] - \
        mesh.vertices[mesh.faces[:, 1], :]

    N = cross_product(e_ij, e_jk)
    A = 0.5 * np.linalg.norm(N, 2, 1)
    A = np.reshape(A, (n_tri, 1))
    N = 1 / (2 * A) * N  # may raise an error or be wrong,
    # careful with dims of A and N

    grad_triangle = (
        texture[mesh.faces[:, 0]] * e_jk
        + texture[mesh.faces[:, 1]] * e_ki
        + texture[mesh.faces[:, 2]] * e_ij
    )

    grad_triangle = 1 / (2 * A) * cross_product(N, grad_triangle)

    # From faces to vertices,
    # use the Nvertex x Ntriangles sparse matrix correspondance
    grad_vertex = mesh.faces_sparse * grad_triangle
    grad_vertex = grad_vertex * \
        np.reshape(1 / mesh.vertex_degree, (n_vertex, 1))

    return grad_vertex


def gradient(mesh, texture_array):
    """
    Compute gradient on a triangular mesh with a scalar function.
    Gradient is computed on each triangle by the function described in
    http://dgd.service.tu-berlin.de/wordpress/vismathws10/2012/10/
    17/gradient-of-scalar-functions/.
    On each vertex, compute the mean gradient of all triangle with the vertex.
    first version author: Guillaume Vicaigne (Internship 2018)
    :param mesh: Triangular mesh
    :param texture_array: Scalar function on Vertices, numpy array
    :return: Gradient on Vertices
    :rtype: numpy.array (update 16/12/2020, J Lefevre)
    """

    # Initialize Parameters
    vert = mesh.vertices
    l_vert = len(vert)
    poly = mesh.faces
    l_poly = len(poly)
    n = 0

    # Initialize Dictionnary
    dicgrad = dict()
    for i in range(l_vert):
        dicgrad[i] = [0, 0, 0, 0]

    # Calculate the Gradient, TO DO: avoid the loop and do only array
    # operations
    gradient_vector = np.zeros((l_vert, 3))
    for i in range(l_poly):
        # Percentage done
        if int(i / float(l_poly) * 100) > n:
            n += 1
            print(str(n) + " %")
        j = []
        for jj in poly[i]:
            j.append(jj)
        grad = [0.0, 0.0, 0.0, 0.0]
        eij = [
            vert[j[1]][0] - vert[j[0]][0],
            vert[j[1]][1] - vert[j[0]][1],
            vert[j[1]][2] - vert[j[0]][2],
        ]
        eki = [
            vert[j[0]][0] - vert[j[2]][0],
            vert[j[0]][1] - vert[j[2]][1],
            vert[j[0]][2] - vert[j[2]][2],
        ]
        ejk = [
            vert[j[2]][0] - vert[j[1]][0],
            vert[j[2]][1] - vert[j[1]][1],
            vert[j[2]][2] - vert[j[1]][2],
        ]
        A = 0.5 * np.linalg.norm(np.cross(eij, ejk))
        N = 0.5 / A * np.cross(ejk, eki)
        grad[0:3] = np.cross(
            0.5 * N / A,
            np.multiply(texture_array[j[0]], ejk)
            + np.multiply(texture_array[j[1]], eki)
            + np.multiply(texture_array[j[2]], eij),
        )
        grad[3] = 1.0
        for jj in j:
            dicgrad[jj] = np.add(dicgrad[jj], grad)
    for i in range(l_vert):
        gradient_vector[i] = np.multiply(dicgrad[i][0:3], 1 / dicgrad[i][3])
    return gradient_vector


def norm_gradient(mesh, texture_array):
    """
    Compute the norm of a vertex Gradient on vertex
    first version author: Guillaume Vicaigne (Internship 2018)
    :param mesh: Triangular mesh
    :param texture_array: Scalar function on Vertices, numpy array
    :return: Gradient's Norm
    """

    # Compute the gradient of the Mesh
    grad = gradient(mesh, texture_array)

    return np.linalg.norm(grad, 2, 1)



import numpy as np
from scipy.spatial import Delaunay
from scipy.optimize import newton
import trimesh
from trimesh import creation as tcr
import slam.topology as stop


def generate_hinge(
    n_hinge=3, n_step=50,
        min_coord=-1 / 5, max_coord=1 / 5, regularity="regular"):
    """
    Generate a hinge shaped surface
    :param n_hinge:
    :param n_step:
    :param min_coord:
    :param max_coord:
    :param regularity:
    :return:
    """
    xmin, xmax = [min_coord, max_coord]
    ymin, ymax = [min_coord, max_coord]

    # Coordinates
    x = np.linspace(xmin, xmax, n_step)
    y = np.linspace(ymin, ymax, n_step)
    X, Y = np.meshgrid(x, y)

    # Delaunay triangulation
    X = X.flatten()
    Y = Y.flatten()
    faces_tri = Delaunay(np.vstack((X, Y)).T, qhull_options="QJ Qt Qbb")

    # Equation

    def cart2pol(x, y):
        rho = np.sqrt(x**2 + y**2)
        phi = np.arctan2(y, x)
        return rho, phi

    R, Phi = cart2pol(X, Y)

    if regularity == "regular":
        alpha_min = 2
        alpha_max = 4
        amplitude = 3
    else:
        alpha_min = 1 / 2
        alpha_max = 2
        amplitude = 0.5

    exponent = (alpha_max - alpha_min) / 2 * np.cos(n_hinge * Phi) + (
        alpha_min + alpha_max
    ) / 2
    Z = -amplitude * R**exponent
    coords = np.stack([X, Y, Z])
    mesh = trimesh.Trimesh(
        faces=faces_tri.simplices,
        vertices=coords.T,
        process=False)
    return mesh


def quadric(K1, K2):
    """
    compute the Z coordinate of a quadric depending on X and Y coordinates
    :param K1:
    :param K2:
    :return:
    """

    def fonction(x, y):
        return K1 * x**2 + K2 * y**2

    return fonction


def quadric_curv_gauss(K):
    """
    analytical Gaussian curvature of a quadric
    :param K1:
    :param K2:
    :return:
    """
    K1 = K[0]
    K2 = K[1]

    def curv_gauss(x, y):
        num = 4 * (K1 * K2)
        denom = (1 + 4 * K1**2 * x**2 + 4 * K2**2 * y**2) ** 2
        return num / denom

    return curv_gauss


def quadric_curv_mean(K):
    """
    analytical mean curvature of a quadric
    :param K1:
    :param K2:
    :return:
    """
    K1 = K[0]
    K2 = K[1]

    def curv_mean(x, y):
        num = -(
            2 * K2 * (1 + 4 * K1**2 * x**2) + 2 * K1 * (1 + 4 * K2**2 * y**2)
        )
        denom = 2 * (1 + 4 * K1**2 * x**2 + 4 * K2**2 * y**2) ** (3 / 2)

        return num / denom

    return curv_mean


def adaptive_sampling(ymax, K, step):
    """
        sample [-ymax,ymax] such as:
        1. y_{i+1}-y_i =h_i with h_i obtained by a recursive formula (1)
        2. f(y_i) is regularly sampled, where f(x)=K x**2

        => same as
        1. Computing the curvilinear abscissa of x -> Kx**2 and
        sample regularly with parameter step *sqrt(3)/2
        s(x) = [ 2*K*x*sqrt((2*K*x)**2+1) + arcsinh(2*K*x) ]/ (4*K)
        2. Come back in the y domain by inverting the curvilinear abscissa
        (Newton method)

    :param ymax:
    :param K: amplitude of the paraboloid
    :param step: desired sampling step if K =0
    :return:
    """
    # Curvilinear abscisse
    def f(x):
        return (2 * K * x * np.sqrt((2 * K * x) ** 2 + 1) +
                np.arcsinh(2 * K * x)) / (4 * K)

    # Step 1
    curve_length = f(ymax)
    curve_step = np.sqrt(3) / 2 * step  # * np.sqrt(K+1) # Pythagore
    Npoints = int(np.floor(curve_length / curve_step))
    curve_samples = np.arange(0, curve_length, curve_step)

    # Step 2
    y_pos = np.zeros((Npoints + 1,))
    for i in range(Npoints + 1):
        y_pos[i] = newton(lambda x: f(x) - curve_samples[i], curve_samples[i])
    y_pos = np.concatenate([-y_pos[::-1], y_pos[1:]])
    curve_samples = np.concatenate([-curve_samples[::-1], curve_samples[1:]])
    return y_pos, curve_samples


def generate_paraboloid_regular(
    A,
    nstep=50,
    ax=1,
    ay=1,
    random_sampling=False,
    random_distribution_type="gaussian",
    ratio=0.1,
):
    """
    generate a regular paraboloid mesh Z=K*Y^2
    ratio and random_distribution_type parameters are unused if
    random_sampling is set to False
    :param A: amplitude of the paraboloid
    :param nstep: nstepx or the sampling step stepx as a float !
    :param ax: half length of the domain
    :param ay: half width of the domain
    :param random_sampling:
    :param random_distribution_type:
    :param ratio:
    :return:
    """
    # Parameters
    xmin, xmax = [-ax, ax]
    ymax = ay  # ymin, ymax = [-ay, ay]
    # Define the sampling
    if isinstance(nstep, int):
        stepx = (xmax - xmin) / nstep
    else:
        stepx = nstep

    # Coordinates
    x = np.arange(xmin, xmax, stepx)
    # To generate y
    y, curve_samples = adaptive_sampling(ymax, A, stepx)

    X, Y = np.meshgrid(x, y)
    X[::2] += stepx / 2
    X = X.flatten()
    Y = Y.flatten()

    # Random perturbation
    if random_sampling:
        sigma = stepx * ratio  # characteristic size of the mesh * ratio
        nb_vert = len(x) * len(y)
        if random_distribution_type == "gamma":
            theta = (
                np.random.rand(
                    nb_vert,
                )
                * np.pi
                * 2
            )
            mean = sigma
            variance = sigma**2
            radius = np.random.gamma(
                mean**2 / variance, variance / mean, nb_vert)
            X = X + radius * np.cos(theta)
            Y = Y + radius * np.sin(theta)
        elif random_distribution_type == "uniform":
            X = X + np.random.uniform(-1, 1, 100)
            Y = Y + np.random.uniform(-1, 1, 100)
        else:
            X = X + sigma * np.random.randn(
                nb_vert,
            )
            Y = Y + sigma * np.random.randn(
                nb_vert,
            )

    # Delaunay triangulation: be careful, to do on the curvilinear aspects to
    # avoid triangle flips
    Xtmp, S = np.meshgrid(x, curve_samples)
    S = S.flatten()
    # faces_tri = Triangulation(X, S)
    faces_tri = Delaunay(np.vstack((X, S)).T, qhull_options="QJ Qt Qbb")
    # alternative setting? 'Qbb Qc Qz Qj'

    Z = quadric(0, A)(X, Y)
    coords = np.array([X, Y, Z]).transpose()

    paraboloid_mesh = trimesh.Trimesh(
        faces=faces_tri.simplices, vertices=coords, process=False
    )
    # remove the faces having any vertex on the boundary to avoid
    # atypical faces geometry due to Delaunay triangulation in 2D
    # TO DO: same boundary removal as for generate_quadric
    return stop.remove_mesh_boundary_faces(
        paraboloid_mesh, face_vertex_number=1)


"""
older, outdated version
# def generate_quadric(K, nstep=50, ax=1, ay=1, random_sampling=True,
#                      ratio=0.2, random_distribution_type='gaussian'):
#     # Parameters
#     xmin, xmax = [-ax, ax]
#     ymin, ymax = [-ay, ay]
#
#     # Coordinates
#     stepx = (xmax - xmin) / nstep
#     x = np.arange(xmin, xmax, stepx)
#     # x, stepx = np.linspace(xmin, xmax, nstep, retstep=True)
#     stepy = stepx * np.sqrt(3) / 2  # to ensure equilateral faces
#     y = np.arange(ymin, ymax, stepy)
#     # y = np.linspace(ymin, ymax, nstep)
#     X, Y = np.meshgrid(x, y)
#     X[::2] += stepx / 2
#     # Y += np.sqrt(3) / 2
#     X = X.flatten()
#     Y = Y.flatten()
#
#     if random_sampling:
#         sigma = stepx * ratio  # characteristic size of the mesh * ratio
#         nb_vert = len(x) * len(y)
#         if random_distribution_type == 'gamma':
#             theta = np.random.rand(nb_vert, ) * np.pi * 2
#             mean = sigma
#             variance = sigma ** 2
#             radius = \
#               np.random.gamma(mean ** 2 / variance, variance / mean, nb_vert)
#             X = X + radius * np.cos(theta)
#             Y = Y + radius * np.sin(theta)
#         elif random_distribution_type == 'uniform':
#             X = X + np.random.uniform(-1, 1, 100)
#             Y = Y + np.random.uniform(-1, 1, 100)
#         else:
#             X = X + sigma * np.random.randn(nb_vert, )
#             Y = Y + sigma * np.random.randn(nb_vert, )
#
#     faces_tri = Delaunay(np.vstack((X, Y)).T, qhull_options='QJ Qt Qbb')
#     # alternative settings? 'Qbb Qc Qz Qj'
#
#     Z = quadric(K[0], K[1])(X, Y)
#     coords = np.array([X, Y, Z]).transpose()
#
#     quadric_mesh = trimesh.Trimesh(faces=faces_tri.simplices,
#                                    vertices=coords,
#                                    process=False)
#     # remove the faces having any vertex on the boundary to avoid
#     # atypical faces geometry due to Delaunay triangulation in 2D
#    return stop.remove_mesh_boundary_faces(quadric_mesh, face_vertex_number=1)
"""


def generate_quadric(
    K,
    nstep=[int(50), int(50)],
    equilateral=False,
    ax=1,
    ay=1,
    random_sampling=True,
    ratio=0.2,
    random_distribution_type="gaussian",
):
    """
    generate a quadric mesh Z=K1*X^2 + K2*Y^2
    ratio and random_distribution_type parameters are unused if
    random_sampling is set to False
    :param K: list with [K1,K2]
    :param nstep: list with [nstepx,nstepy] or the sampling steps
    [stepx,stepy] as floats !
    :param equilateral: to have an equilateral sampling scheme of the quadric
    :param ax: half length of the domain
    :param ay: half width of the domain
    :param random_sampling:
    :param ratio:
    :param random_distribution_type:
    :return:
    """

    # Parameters
    xmin, xmax = [-ax, ax]
    ymin, ymax = [-ay, ay]
    # Define the sampling
    if equilateral:
        if isinstance(nstep[0], int):
            stepx = (xmax - xmin) / nstep[0]
        else:
            stepx = nstep[0]
        stepy = stepx * np.sqrt(3) / 2  # to ensure equilateral faces
    else:
        if isinstance(nstep[0], int):
            stepx = (xmax - xmin) / nstep[0]
            stepy = (ymax - ymin) / nstep[1]
        else:
            stepx = nstep[0]
            stepy = nstep[1]

    # Coordinates
    x = np.arange(xmin, xmax, stepx)
    # x, stepx = np.linspace(xmin, xmax, nstep, retstep=True)
    y = np.arange(ymin, ymax, stepy)
    # y = np.linspace(ymin, ymax, nstep)
    X, Y = np.meshgrid(x, y)

    # Boundary of a meshgrid-like set of points,
    # warning X.max()/Y.max() and not xmax, ymax
    boundary_x = np.logical_or(X.flatten() == xmin, X.flatten() == X.max())
    boundary_y = np.logical_or(Y.flatten() == ymin, Y.flatten() == Y.max())
    boundary = np.logical_or(boundary_x, boundary_y)
    boundary = np.where(boundary)[0]

    # Adapt in case of equilateral meshing
    if equilateral:
        X[::2] += stepx / 2
    # Y += np.sqrt(3) / 2
    X = X.flatten()
    Y = Y.flatten()

    if random_sampling:
        sigma = stepx * ratio  # characteristic size of the mesh * ratio
        nb_vert = len(x) * len(y)
        if random_distribution_type == "gamma":
            theta = (
                np.random.rand(
                    nb_vert,
                )
                * np.pi
                * 2
            )
            mean = sigma
            variance = sigma**2
            radius = np.random.gamma(
                mean**2 / variance, variance / mean, nb_vert)
            X = X + radius * np.cos(theta)
            Y = Y + radius * np.sin(theta)
        elif random_distribution_type == "uniform":
            X = X + np.random.uniform(-1, 1, 100)
            Y = Y + np.random.uniform(-1, 1, 100)
        else:
            X = X + sigma * np.random.randn(
                nb_vert,
            )
            Y = Y + sigma * np.random.randn(
                nb_vert,
            )

    # Delaunay triangulation, based on scipy binding of Qhull.
    # See https://scipy.github.io/devdocs/generated/scipy.spatial.Delaunay.html
    #     scipy.spatial.Delaunay
    #     and http://www.qhull.org/html/qdelaun.htm for more informations
    faces_tri = Delaunay(np.vstack((X, Y)).T, qhull_options="QJ Qt Qbb")
    # alternative settings? 'Qbb Qc Qz Qj'

    Z = quadric(K[0], K[1])(X, Y)
    coords = np.array([X, Y, Z]).transpose()

    quadric_mesh = trimesh.Trimesh(
        faces=faces_tri.simplices, vertices=coords, process=False
    )

    # Remove boundary, computed previously
    boundary_faces = stop.ismember(quadric_mesh.faces, boundary)
    # compute the mask of faces to keep
    # faces for which face_vertex_number or more vertices are on the boundary
    # are excluded from the mask.
    # i.e. faces with 3-face_vertex_number vertices on the boundary are kept
    face_mask = np.logical_not(np.sum(boundary_faces, 1) >= 3)
    quadric_mesh.update_faces(face_mask)
    quadric_mesh.remove_unreferenced_vertices()

    return quadric_mesh


def generate_ellipsiod(a, b, nstep, random_sampling=False):
    """
    generate an ellipsoid
    :param a:
    :param b:
    :param nstep:
    :param random_sampling:
    :return:
    """
    # Coordinates
    if random_sampling:
        THETA = (np.random.rand(nstep * nstep, 1) - 1 / 2) * np.pi
        PHI = 2 * np.pi * np.random.rand(nstep * nstep, 1)
    else:
        theta = np.linspace(-np.pi / 2, np.pi / 2, nstep)
        phi = np.linspace(0, 2 * np.pi, nstep)
        THETA, PHI = np.meshgrid(theta, phi)

    # Sphere coordinates
    X = a * np.cos(THETA) * np.cos(PHI)
    Y = b * np.cos(THETA) * np.sin(PHI)
    Z = np.sin(THETA)

    coords = np.array([X, Y, Z]).squeeze().transpose()

    return tri_from_hull(coords)


def tri_from_hull(vertices):
    """
    compute faces from vertices using trimesh convex hull
    :param vertices: (n, 3) float
    :return:
    """
    mesh = trimesh.Trimesh(vertices=vertices, process=False)
    return mesh.convex_hull


def generate_sphere_random_sampling(vertex_number=100, radius=1.0):
    """
    generate a sphere with random sampling
    :param vertex_number: number of vertices in the output spherical mesh
    :param radius: radius of the output sphere
    :return:
    """
    coords = np.zeros((vertex_number, 3))
    for i in range(vertex_number):
        M = np.random.normal(size=(3, 3))
        Q, R = np.linalg.qr(M)
        coords[i, :] = Q[:, 0].transpose() * np.sign(R[0, 0])
    if radius != 1:
        coords = radius * coords
    return tri_from_hull(coords)


def generate_sphere_icosahedron(subdivisions=3, radius=1.0):
    """
    generate a sphere by subdividing an icosahedron
    simply call the trimesh function
    see trimesh.creation.icosphere for more details
    :param subdivisions:  int
      How many times to subdivide the mesh.
      Note that the number of faces will grow as function of
      4 ** subdivisions, so you probably want to keep this under ~5
    :param radius: float
      Desired radius of sphere
    :return:
    """
    return tcr.icosphere(subdivisions=subdivisions, radius=radius)


def compute_weingarten_map(K, x, y):
    """
    compute the weingarten matrix of a quadric depending on x and y coordinates
    :param K: (2,) array such as z(x,y)=K[0]*x**2 + K[1]*y**2
    :param (x,y): coordinates where to compute the matrix
    :return: (2,2) matrix of the weingarten endomorphism to be evaluated at
    (x,y)
    """
    K1 = K[0]
    K2 = K[1]
    M = np.zeros((2, 2), dtype=float)
    coeff = 4 * K1 * K2 * x * y
    M[0, 0] = K1 * (1 + 4 * K2**2 * y**2)
    M[0, 1] = -coeff * K2
    M[1, 0] = -coeff * K1
    M[1, 1] = K2 * (1 + 4 * K1**2 * x**2)
    return M


def compute_principal_directions(K, x, y):
    """
    Compute the principal direction of a quadric, obtained as eigenvectors of
    the Weingarten matrix
    :param K: (2,) array such as z(x,y)=K[0]*x**2 + K[1]*y**2
    :param x: x coordinate where to compute the matrix
    :param y: y coordinate where to compute the matrix
    :return: a 2-list of (2,1) vectors, i.e. principal directions ordered with
    Kmin,Kmax respectively
    """
    M = compute_weingarten_map(K, x, y)
    w, v = np.linalg.eig(M)
    if w[0] > w[1]:
        return v[:, 1], v[:, 0]
    return v[:, 0], v[:, 1]


def compute_local_basis(K, x, y):
    """
    Compute the local basis of the tangent plane for the quadric
    z(x,y)=K[0]*x**2+K[1]*y**2
    :param K: (2,) array such as z(x,y)=K[0]*x**2 + K[1]*y**2
    :param x: x coordinate where to compute the matrix
    :param y: y coordinate where to compute the matrix
    :return: a 2-list of (3,1) vectors, obtained as
    d (x,y,z(x,y) /dx and d (x,y,z(x,y) /dy
    """
    e1 = np.array([[1, 0, 2 * K[0] * x]])
    e2 = np.array([[0, 1, 2 * K[1] * y]])
    return e1, e2


def compute_all_principal_directions(K, vertices):
    """
    Compute all the principal directions of a quadric surface defined through
    the input K and sampled points vertices in a local basis
    :param K: (2,) array such as z(x,y)=K[0]*x**2 + K[1]*y**2
    :param vertices: (n,3) coordinates of the quadric. It is implicitly assumed
    that vertices[:,0] and vertices[:,1]
    corresponds to x and y axis
    :return: res: a (n,2,2) array such as res[i,:,0] and res[i,:,1] are the two
    principal directions at vertex vertices[i,:] in the local basis
    """
    n = vertices.shape[0]
    res = np.zeros((n, 2, 2), dtype=float)
    for i in range(n):
        u1, u2 = compute_principal_directions(
            K, vertices[i, 0], vertices[i, 1])
        res[i, :, 0] = u1
        res[i, :, 1] = u2
    return res


def compute_all_principal_directions_3D(K, vertices):
    """
    Compute all the principal directions of a quadric surface defined through
    the input K and sampled points vertices in 3D
    :param K: (2,) array such as z(x,y)=K[0]*x**2 + K[1]*y**2
    :param vertices: (n,3) coordinates of the quadric. It is implicitly assumed
    that vertices[:,0] and vertices[:,1]
    corresponds to x and y axis
    :return: res: a (n,3,2) array such as res[i,:,0] and res[i,:,1] are the two
    principal directions at vertex vertices[i,:]
    """
    n = vertices.shape[0]
    res = np.zeros((n, 3, 2), dtype=float)
    for i in range(n):
        u1, u2 = compute_principal_directions(
            K, vertices[i, 0], vertices[i, 1])
        e1, e2 = compute_local_basis(K, vertices[i, 0], vertices[i, 1])
        res[i, :, 0] = u1[0] * e1 + u1[1] * e2
        res[i, :, 1] = u2[0] * e1 + u2[1] * e2
    return res



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
from setuptools import setup, find_packages

BASE_REQUIREMENTS = ["numpy", "scipy", "trimesh", "nibabel", "networkx"]
TEST_REQUIREMENTS = ["flake8", "autopep8", "pytest", "pytest-cov", "coveralls"]

DOC_REQUIREMENTS = ['sphinx',
                    'sphinx-gallery',
                    'sphinx_bootstrap_theme',
                    'numpydoc',
                    'six',
                    'python-dateutil',
                    'sphinxcontrib-fulltoc',
                    'matplotlib']

DIST = ["tvb-gdist"]

TRIMESH_FULL = ["rtree", "shapely"]

# grab version
verstr = "unknown"
try:
    verstrline = open('slam/_version.py', "rt").read()
except EnvironmentError:
    pass  # Okay, there is no version file.
else:
    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    mo = re.search(VSRE, verstrline, re.M)
    if mo:
        verstr = mo.group(1)
    else:
        raise RuntimeError("unable to find version in yourpackage/_version.py")

setup(
    name="brain-slam",
    version=verstr,
    packages=find_packages(),
    author="Guillaume Auzias",
    description="Surface anaLysis And Modeling",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/brain-slam/slam",
    license="MIT",
    python_requires=">=3.12",  # enforce Python 3.12 as minimum
    install_requires=BASE_REQUIREMENTS,
    extras_require={
        "full": DIST + TRIMESH_FULL,
        "dev": DIST + TEST_REQUIREMENTS + TRIMESH_FULL,
        "doc": DIST + TEST_REQUIREMENTS + TRIMESH_FULL + DOC_REQUIREMENTS,
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
    ],
)



from slam import generate_parametric_surfaces as sps


def test_generate_ellipsoid():
    nstep = 50
    randomSampling = True
    a = 2
    b = 1
    ellips = sps.generate_ellipsiod(a, b, nstep, randomSampling)
    assert len(ellips.vertices) == nstep**2



import numpy as np
import gdist
import networkx as nx


def shortest_path(mesh, start_idx, end_idx):
    # edges without duplication
    edges = mesh.edges_unique

    # the actual length of each unique edge
    length = mesh.edges_unique_length

    # create the graph with edge attributes for length
    # g = nx.Graph()
    # for edge, L in zip(edges, length):
    #     g.add_edge(*edge, length=L)

    # alternative method for weighted graph creation
    # you can also create the graph with from_edgelist and
    # a list comprehension, which is like 1.5x faster

    ga = nx.from_edgelist([(e[0], e[1], {"length": L})
                           for e, L in zip(edges, length)])

    # run the shortest path query using length for edge weight
    path = nx.shortest_path(
        ga,
        source=start_idx,
        target=end_idx,
        weight="length")

    return path


def compute_gdist(mesh, source_id, target_id=None):
    """
    This func computes the geodesic distance between a set
    of sources and targets on a mesh surface by using the
    gdist.compute_gdist().
    If no target_id are provided, all vertices of the mesh
    are considered as targets.

    :param mesh: (trimesh object) the mesh surface
    :param source_id: (list) the sources index
    :param target_id: (list) the targets index
    :return:
    """
    vert = mesh.vertices
    poly = mesh.faces.astype(np.int32)

    source_index = np.array([source_id], dtype=np.int32)
    if target_id:
        target_index = np.array([target_id], dtype=np.int32)
    else:
        target_index = np.linspace(0, len(vert) - 1,
                                   len(vert)).astype(np.int32)

    return gdist.compute_gdist(vert, poly, source_index, target_index)


def local_gdist_matrix(mesh, max_geodist):
    """
    For every vertex, get all the vertices within the maximum distance.
    NOTE: It will take some time to compute all the geo-distance from point to
     point.
    Details about the computing time and memory see in method
    gdist.local_gdist_matrix
    :param mesh:
    :param max_geodist:
    :return:
    """

    vert = mesh.vertices
    poly = mesh.faces.astype(np.int32)

    return gdist.local_gdist_matrix(vert, poly, max_geodist)
